import asyncio
import ipaddress
import logging
import re
import socket
import struct
import time
from collections import defaultdict
from typing import Any, Optional, Tuple, Union, cast

from .network import AF_PACKET, getifaddrs

# Art-Net implementation for Python asyncio
# Any page references to 'spec' refer to
#   "Art-Net 4 Protocol Release V1.4 Document Revision 1.4di 29/7/2023"
#
# This implementation Copyright Tea Engineering Ltd. 2024
#

# Art-Net specific constants
ARTNET_PORT = 6454
ARTNET_PREFIX = bytes("Art-Net".encode() + b"\000")

DMX_UNIVERSE_SIZE = 512

# We need to interrogate network interfaces to check which are configured for IP and
# have a valid broadcast address. For unix-like systems this is fone with socket fnctl
# calls. This isn't portable and might need reconsidering later. Constants from
# https://elixir.bootlin.com/linux/v4.2/source/include/uapi/linux/sockios.h#L43
SIOCGIFADDR = 0x8915
SIOCGIFNETMASK = 0x891B
SIOCGIFBRDADDR = 0x8919
SIOCGIFFLAGS = 0x8913

# interfaces *with an ip* are preferred in this order
PREFERED_INTERFACES_ORDER = ["enp.*", "wlp.*"]

# register a logger so that our debug can be enabled if required
logger = logging.getLogger("aioartnet")


# helper to de-tangle some of the protocol endianness. Fields like IP address
# are stored as 4 consecutive bytes, but not in little-endian like the rest of the
# protocol. Better to read as a 32-bit int in struct.unpack and then byteswap it
def swap32(x: int) -> int:
    return int.from_bytes(
        x.to_bytes(4, byteorder="little"), byteorder="big", signed=False
    )


def swap16(x: int) -> int:
    return int.from_bytes(
        x.to_bytes(2, byteorder="little"), byteorder="big", signed=False
    )


# The broadcast IP is used for locating nodes and managing subscriptions.
# Art-Net I and II protocols also used it for sending DMX data, and many
# implementations are backwards compatible.
# Art-Net II switched over to UDP unicast from the universe publisher to
# the subscriber(s).

# We enumerate all universes offered by all nodes and offer them
# merged by the 15-bit universe identifier. Controlling code can make a call to
# set_port_config to subscribe, write, or broadcast each universe, and we will
# manage the art-poll-reply flags to make this work.
#
# We might recieved unsolicited broadcasts of a universe from other controllers,
# (like QLC+) there's nothing we can do about this, but we drop them unless we
# are in subscribe mode.

# Each Node can publish one (artnet<3) or more (arcnet>3) sets of 4-ports.
# Each set fixes a single net and sub_net value, however the choice
# of universe nibble is determined per-port (net:sub_net:universe).

DGAddr = tuple[Union[str, Any], int]


class ArtNetNode:
    def __init__(
        self,
        longName: str,
        portName: str,
        style: int,
        ip: str,
        udpport: int,
    ) -> None:
        self.portName = portName
        self.longName = longName
        self._portBinds: defaultdict[int, list[ArtNetPort]] = defaultdict(list)
        self.ports: list[ArtNetPort] = []
        self.style: int = style
        self.udpport = udpport
        self.ip = ip
        self.last_reply: float = 0.0

    def __repr__(self) -> str:
        return f"ArtNetNode<{self.portName},{self.ip}:{self.udpport}>"


class ArtNetUniverse:
    def __init__(self, portaddress: int, client: "ArtNetClient"):
        if portaddress > 0x7FFF:
            raise ValueError("Invalid net:subnet:universe, as net>128")
        self.portaddress = portaddress
        self.publishers: list[ArtNetNode] = list()
        self.subscribers: list[ArtNetNode] = list()
        self.last_data = bytearray(DMX_UNIVERSE_SIZE)
        self._last_seq = 1
        self._last_publish: float = 0.0
        self.publisherseq: dict[Tuple[DGAddr, int], int] = {}
        self._client = client

    def split(self) -> Tuple[int, int, int]:
        # name  net:sub_net:universe
        # bits  8:15  4:8     0:4
        net = self.portaddress >> 8
        sub_net = (self.portaddress >> 4) & 0x0F
        universe = self.portaddress & 0x0F
        return net, sub_net, universe

    def __repr__(self) -> str:
        net, sub_net, universe = self.split()
        return f"{net}:{sub_net}:{universe}"

    def set_dmx(self, data: bytes) -> None:
        assert len(data) == DMX_UNIVERSE_SIZE
        self.last_data[:] = data[:]
        self._client._send_dmx(self)

    def get_dmx(self) -> bytes:
        return self.last_data


class ArtNetPort:
    def __init__(
        self,
        node: Union[ArtNetNode, "ArtNetClient"],
        is_input: bool,
        media: int,
        portaddr: int,
        universe: ArtNetUniverse,
    ):
        self.node = node
        self.is_input = is_input
        self.media = 0
        self.portaddr = portaddr
        self.universe = universe

    def __repr__(self) -> str:
        inout = {True: "Input", False: "Output"}[self.is_input]
        media = ["DMX", "MIDI", "Avab", "Colortran CMX", "ADB 62.5", "Art-Net", "DALI"][
            self.media
        ]
        return f"Port<{inout},{media},{self.universe}>"


class ArtNetClientProtocol(asyncio.DatagramProtocol):
    def __init__(self, client: "ArtNetClient"):
        self.client = client
        self.transport: Optional[asyncio.DatagramTransport] = None
        self._last_poll = 0.0
        self.handlers = {
            0x2000: self.on_art_poll,
            0x2100: self.on_art_poll_reply,
            0x5000: self.on_art_dmx,
        }
        client.protocol = self
        self.node_report_counter = 0

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        self.transport = cast(asyncio.DatagramTransport, transport)
        sock = transport.get_extra_info("socket")
        if sock:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    def datagram_received(self, data: bytes, addr: DGAddr) -> None:
        if data[0:8] == ARTNET_PREFIX:
            (opcode,) = struct.unpack("H", data[8:10])
            h = self.handlers.get(opcode, None)
            if h:
                h(addr, data[10:])
            else:
                logger.debug(
                    f"Received unsupported Art-Net: op {hex(opcode)} from {addr}: {data[10:]!r}"
                )
        else:
            logger.debug(f"Received non Art-Net data from {addr}: {data!r}")

    def on_art_poll(self, addr: DGAddr, data: bytes) -> None:
        ver, flags, priority = struct.unpack("<HBB", data)
        ver = swap16(ver)
        logger.debug(
            f"Received Art-Net Poll: ver {ver} flags {flags} prio: {priority} from {addr}"
        )
        self.send_art_poll_reply()

    def on_art_poll_reply(self, addr: DGAddr, data: bytes) -> None:
        # everything up to the mac address field is mandatory, the rest must be
        # parsed only if it is sent (field at a time)
        ip, port, fw, netsw, subsw, oemCode = struct.unpack("<IHHBBH", data[0:12])
        ubeaVer, status, esta, portName = struct.unpack("<BBH18s", data[12:34])
        longName, report, numports = struct.unpack("<64s64sH", data[34:164])
        ptype, ins, outs, swin, swout = struct.unpack("<4s4s4s4s4s", data[164:184])
        acnprio, swmacro, swremote, style = struct.unpack("<BBB3xB", data[184:191])
        # mac = struct.unpack("<6s", data[191:197])

        bindip = 0
        bindindex = 0
        status2 = 0
        goodout = bytes(4)
        status3 = 0
        # rdm = bytes(6)
        user = 0
        refresh = 0
        if len(data) >= 202:
            bindip, bindindex = struct.unpack("<IB", data[197:202])
        if len(data) >= 203:
            (status2,) = struct.unpack("<B", data[202:203])
        if len(data) >= 207:
            (goodout,) = struct.unpack("<4s", data[203:207])
        if len(data) >= 208:
            (status3,) = struct.unpack("<B", data[207:208])
        if len(data) >= 216:
            (rdm,) = struct.unpack("<6s", data[210:216])
        if len(data) >= 218:
            (user,) = struct.unpack("<H", data[216:218])
        if len(data) >= 220:
            (refresh,) = struct.unpack("<H", data[218:220])

        # post process
        portName = portName.rstrip(b"\000").decode()
        longName = longName.rstrip(b"\000").decode()

        ipa = ipaddress.IPv4Address(swap32(ip))

        nn = self.client.nodes.get(ip, None)
        changed = False
        if nn is not None:
            changed |= nn.longName != longName
            changed |= nn.portName != portName
            changed |= nn.style != style
            logger.debug(f"change detection on {nn} => {changed}")

        # FIXME: what if a node changes ip address?

        if nn is None or changed:
            newnode = ArtNetNode(
                ip=f"{ipa}",
                udpport=port,
                longName=longName,
                portName=portName,
                style=style,
            )
            if changed:
                logger.debug(f"change detected: from {nn} to {newnode}")

            # TODO: make client.node observable dict and remove this method
            self.client.add_node(ip, newnode)
            nn = newnode

        nn.last_reply = time.time()

        # iterate through the ports and create ports and universes
        portList = []
        for _type, _in, _out, _swin, _swout in zip(ptype, ins, outs, swin, swout):
            in_port_addr = (
                ((netsw & 0x7F) << 8) + ((subsw & 0x0F) << 4) + (_swin & 0x0F)
            )
            out_port_addr = (
                ((netsw & 0x7F) << 8) + ((subsw & 0x0F) << 4) + (_swout & 0x0F)
            )
            if _type & 0b10000000:
                outu = self.client._get_create_universe(out_port_addr)
                portList.append(
                    ArtNetPort(nn, False, _type & 0x1F, out_port_addr, outu)
                )
            if _type & 0b01000000:
                inu = self.client._get_create_universe(in_port_addr)
                portList.append(ArtNetPort(nn, True, _type & 0x1F, in_port_addr, inu))

        # track which 'pages' of port bindings we have seen
        old_ports = nn._portBinds[bindindex]
        for p in portList:
            if p not in old_ports:
                nn.ports.append(p)
                nn._portBinds[bindindex].append(p)
                {True: p.universe.publishers, False: p.universe.subscribers}[
                    p.is_input
                ].append(nn)

        for p in list(old_ports):
            if p not in portList:
                nn.ports.remove(p)
                nn._portBinds[bindindex].remove(p)
                {True: p.universe.publishers, False: p.universe.subscribers}[
                    p.is_input
                ].remove(nn)
        logger.debug(
            f"Received Art-Net PollReply from {ip} fw {fw} portName {portName} longName: {longName} bindindex {bindindex} ports:{portList}"
        )

    def on_art_dmx(self, addr: DGAddr, data: bytes) -> None:
        ver, seq, phys, sub, net, chlen = struct.unpack("<HBBBBH", data[0:8])
        ver = swap16(ver)
        chlen = swap16(chlen)
        portaddress = sub + (net << 8)

        # check enabled flag as a ~40Hz frequency message
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(
                f"Received Art-Net DMX: ver {ver} port_address {portaddress} seq {seq} channels {chlen} from {addr}"
            )
        # TODO: should we support overside/undersize universes? For now truncate
        # what is supplied beyond DMX_UNIVERSE_SIZE
        if chlen > DMX_UNIVERSE_SIZE:
            chlen = DMX_UNIVERSE_SIZE

        u = self.client._get_create_universe(portaddress)

        # If addr is the broadcast address, it could be from a device
        # that is not responding to Art-Net Poll messages, rather then a unicast
        # stream just to us.

        # It would seem that (ip_address,phys) is the 'key' for a sequence.
        # ie. the spec envisions that a single node with two input ports could output
        # them both on Art-Net to the same universe. The reciever would disambiguate
        # the senders by ip_address+phys (see p62)
        publisherkey = (addr, phys)

        # TODO: check seq field for packet re-ordering and loss detection
        # a seq of 0 is considered always 'in sequence' as the sender is not
        # supporting sequencing. Valid values wrap within [1..255]
        if seq > 0:
            # store the last value we've seen
            u.publisherseq[publisherkey] = seq

        # TODO: HTP/LTP merging with Merge Mode, see "Data Merging" spec p61
        # See ArtAddress AcCancelMerge flags spec p39
        # Only two sources are allowed to contribute to the values in the universe
        u.last_data[0:chlen] = data[8 : 8 + chlen]

    async def art_poll_task(self) -> None:
        while True:
            await asyncio.sleep(0.1)
            t = time.time()

            for u in self.client._publishing:
                if t > u._last_publish + 1.0:
                    self._send_art_dmx(u)

            if t > self._last_poll + 2.0:
                self._send_art_poll()

    def _send_art_poll(self) -> None:
        self._last_poll = time.time()
        self.node_report_counter = (self.node_report_counter + 1) % 10000
        message = ARTNET_PREFIX + struct.pack("<HBBBB", 0x2000, 0, 14, 2, 16)
        logger.debug(f"sending poll to {self.client.broadcast_ip}")
        if self.transport:
            self.transport.sendto(message, addr=(self.client.broadcast_ip, ARTNET_PORT))

    def _send_art_dmx(self, u: ArtNetUniverse) -> None:
        u._last_publish = time.time()
        u._last_seq = (u._last_seq + 1) % 255
        logger.debug(f"send_art_dmx {u} to {u.subscribers}")
        for s in u.subscribers:
            self._send_art_dmx_subscriber(u, s, u._last_seq)

    def send_art_poll_reply(self) -> None:
        for bi, p in self.client._portBinds.items():
            self._send_art_poll_reply_bindindex(bi, p)

    def _send_art_poll_reply_bindindex(
        self, bindindex: int, ports: list[ArtNetPort]
    ) -> None:
        if not self.client.unicast_ip:
            return
        ip = int.from_bytes(
            socket.inet_aton(self.client.unicast_ip), byteorder="little", signed=False
        )
        bindip = ip
        oemCode = 0x2CD3
        esta = 0x02AE
        status2 = 0x08  # 15-bit port-address supported
        status3 = 0
        user = 0
        refresh = 0

        # build dynamically from ports[]
        ptype = bytearray(4)
        ins = bytearray(4)
        outs = bytearray(4)
        swin = bytearray(4)
        swout = bytearray(4)
        goodout = bytearray(4)
        # defaults used when we have zero ports!
        net = self.client.net
        subnet = self.client.subnet
        mac = self.client.mac

        for i, p in enumerate(ports):
            ptype[i] = p.media | (1 << (6 if p.is_input else 7))
            net, subnet, universe = p.universe.split()
            if p.is_input:
                swin[i] = universe
            else:
                swout[i] = universe

        nodereport = f"#0001 [{self.node_report_counter:04d}] Debug OK".encode()

        data = ARTNET_PREFIX + struct.pack(
            "<HIHHBBHBBH18s64s64sBB4s4s4s4s4sBBB3xB6sIBB4sB6xHH11x",
            0x2100,
            ip,
            ARTNET_PORT,
            1,
            net,
            subnet,
            oemCode,
            0,
            0,
            esta,
            self.client._portName.encode(),
            self.client._longName.encode(),
            nodereport,
            0,
            len(ports),
            ptype,
            ins,
            outs,
            swin,
            swout,
            0,
            0,
            0,
            self.client._style,
            mac,
            bindip,
            bindindex,
            status2,
            goodout,
            status3,
            user,
            refresh,
        )
        addr = (self.client.broadcast_ip, ARTNET_PORT)
        logger.debug(f"sending poll reply to {addr}")
        if self.transport:
            self.transport.sendto(data, addr=addr)

    def _send_art_dmx_subscriber(
        self, universe: ArtNetUniverse, node: ArtNetNode, seq: int
    ) -> None:
        subuni = universe.portaddress & 0xFF
        net = universe.portaddress >> 8
        message = ARTNET_PREFIX + struct.pack(
            "<HBBBBBBH",
            0x5000,
            0,
            14,
            1 + seq,
            0,
            subuni,
            net,
            swap16(DMX_UNIVERSE_SIZE),
        )
        message = message + universe.last_data

        logger.debug(
            f"sending dmx for {universe} to {node} at {node.ip}:{node.udpport}"
        )
        if self.transport:
            self.transport.sendto(message, addr=(node.ip, node.udpport))

    def error_received(self, exc: Exception) -> None:
        logger.warn("Error received:", exc)

    def connection_lost(self, exc: Optional[Exception]) -> None:
        if self.client._task:
            logger.info("Connection closed, cancelling periodic task")
            self.client._task.cancel()


UniverseKey = Union[int, str, ArtNetUniverse]


class ArtNetClient:
    def __init__(
        self,
        interface: Optional[str] = None,
        net: int = 0,
        subnet: int = 0,
        passive: bool = False,
        portName: str = "aioartnet",
    ) -> None:
        self.nodes: dict[int, ArtNetNode] = {}
        self.universes: dict[int, ArtNetUniverse] = {}
        self.net = 0
        self.subnet = 0
        self.ports: list[ArtNetPort] = []
        self._portBinds: dict[int, list[ArtNetPort]] = {1: []}
        self._portName = portName
        self._longName = f"{portName} (aioartnet)"
        # identify as an StController, see Table 4 'Style Codes' p23
        self._style = 1
        self.passive = passive
        self.broadcast_ip: Optional[str] = None
        self.unicast_ip: Optional[str] = None
        self.listen_ip: str = "0.0.0.0"
        self.mac: bytes = b"\01\22\33\44\55\66"

        self.protocol: Optional[ArtNetClientProtocol] = None
        self._publishing: list[ArtNetUniverse] = []
        self.interface: Optional[str] = interface
        self._task: Optional[asyncio.Task[None]] = None

    async def connect(self) -> asyncio.DatagramTransport:
        loop = asyncio.get_running_loop()

        if self.broadcast_ip is None or self.unicast_ip is None:
            if self.interface is None:
                self.interface = get_preferred_artnet_interface()

            ips = get_iface_ip(self.interface)
            if not ips:
                raise Exception(
                    f"Unable to determine IPs for interface '{self.interface}'"
                )
            self.unicast_ip = ips["addr"]
            self.broadcast_ip = ips["broadaddr"]
            self.mac = bytes.fromhex(ips["mac"])

        logger.info(
            f"using interface {self.interface} with ip {self.unicast_ip} broadcast ip {self.broadcast_ip}, listening on {self.listen_ip} our mac {self.mac.hex()}"
        )

        # create_datagram_endpoint only allows us to listen on a single address. Ideally we would listen for packets
        # sent to both our unicast and the broadcast address. As a workaround we can listen on 0.0.0.0 which will get
        # us both (however perhaps also those from other interfaces).
        transport, protocol = await loop.create_datagram_endpoint(
            lambda: ArtNetClientProtocol(self),
            local_addr=(self.listen_ip, ARTNET_PORT),
            family=socket.AF_INET,
            allow_broadcast=True,
        )
        if not self.passive:
            self._task = asyncio.create_task(
                protocol.art_poll_task(), name="Art-Net timer"
            )

        return transport

    def set_dmx(self, universe: UniverseKey, data: bytes) -> None:
        assert len(data) == DMX_UNIVERSE_SIZE
        port_addr = self._parse_universe(universe)

        # where to keep our write buffer?
        u = self._get_create_universe(port_addr)
        u.set_dmx(data)

    def _send_dmx(self, universe: ArtNetUniverse) -> None:
        if universe not in self._publishing:
            raise ValueError(f"No input port configured for {universe}")

        if self.protocol:
            self.protocol._send_art_dmx(universe)

    def get_nodes(self) -> list[ArtNetNode]:
        return list(self.nodes.values())

    def add_node(self, ip: int, node: ArtNetNode) -> None:
        self.nodes[ip] = node

    def _parse_universe(self, universe: UniverseKey) -> int:
        if isinstance(universe, str):
            # parse to int
            net, sub, univ = map(int, universe.split(":"))
            return ((net & 0x7F) << 8) + ((sub & 0x0F) << 4) + (univ & 0x0F)
        elif isinstance(universe, int):
            if universe > 0x7FFF:
                raise ValueError(f"invalid universe: {universe} exceeds 0x7fff")
            return universe
        elif isinstance(universe, ArtNetUniverse):
            return universe.portaddress
        else:
            raise ValueError(f"invalid universe: {universe}")

    def set_port_config(
        self,
        universe: UniverseKey,
        is_input: bool = False,
        is_output: bool = False,
    ) -> ArtNetUniverse:
        port_addr = self._parse_universe(universe)

        u = self._get_create_universe(port_addr)

        # port objects within client.ports all have node=None, this is the template
        # for the publisher. client.nodes[ownip].ports[] *should* contain the same
        # information once we process our own replies.
        port = None
        for i, p in enumerate(self.ports):
            if p.universe == u:
                logger.debug(f"set_port_config: has existing port {p}")
                port = p
                break
        if port:
            self.ports.remove(port)
            logger.info(f"removed own port {port}")

        if is_input or is_output:
            port = ArtNetPort(
                node=self, is_input=is_input, media=0, portaddr=port_addr, universe=u
            )
            self.ports.append(port)
            logger.info(f"configured own port {port}")

        # TODO: optimise the layour of self.ports to self._portBinds
        # up to four ports with a common (net,sub-net) can be listed on the same page
        # ie. sharing the same bindIndex. For now, we will put one on each
        if self.ports:
            self._portBinds = dict([(i + 1, [p]) for i, p in enumerate(self.ports)])
        else:
            self._portBinds = {1: []}

        # used for the timer-based DMX repeating
        if u in self._publishing:
            self._publishing.remove(u)
        if is_input:
            self._publishing.append(u)

        if not self.passive and self.protocol:
            self.protocol.send_art_poll_reply()

        return u

    def _get_create_universe(self, port_addr: int) -> ArtNetUniverse:
        if (u := self.universes.get(port_addr, None)) is None:
            u = ArtNetUniverse(port_addr, self)
            self.universes[port_addr] = u
        return u

    # MUTABLE PROPERTIES
    # if you need to change a lot of properties in bulk, set client.passive=True,
    # make the required changes, then clear .passive
    @property
    def portName(self) -> str:
        return self._portName

    @portName.setter
    def portName(self, value: str) -> None:
        self._portName = value
        if not self.passive and self.protocol:
            self.protocol.send_art_poll_reply()

    @property
    def longName(self) -> str:
        return self._longName

    @longName.setter
    def longName(self, value: str) -> None:
        self._longName = value
        if not self.passive and self.protocol:
            self.protocol.send_art_poll_reply()

    @property
    def style(self) -> int:
        return self._style

    @style.setter
    def style(self, value: int) -> None:
        self._style = value
        if not self.passive and self.protocol:
            self.protocol.send_art_poll_reply()


def get_iface_ip(iface: str) -> Optional[dict[str, Any]]:
    """
    Get network interface IP using the network interface name
    :param iface: Interface name (like eth0, enp2s0, etc.)
    :return IP address in the form XX.XX.XX.XX
    """
    na = {}
    for netaddr in getifaddrs(family=socket.AF_INET, ifname=iface):
        # return first address if multiple bound to NIC
        na = dict(**netaddr)
    for netaddr in getifaddrs(family=AF_PACKET, ifname=iface):
        # return first address if multiple bound to NIC
        na["mac"] = netaddr["addr"]
    return na


def get_preferred_artnet_interface() -> str:
    preferred = []
    patterns = list(map(re.compile, PREFERED_INTERFACES_ORDER))

    for netaddr in getifaddrs(family=socket.AF_INET):
        name = str(netaddr["name"])
        packet = netaddr["addr"]
        netmask = netaddr["netmask"]
        bcast = netaddr["broadaddr"]
        logger.debug(f"interface name={name} {packet} {netmask} {bcast}")
        # looks like an explicit class-A primary interface for Art-Net
        if netmask == "255.0.0.0" and packet.startswith("2."):
            preferred.append((-1, name))
        else:
            for i, pattern in enumerate(patterns):
                if re.match(pattern, name):
                    preferred.append((i, name))
                    break
            else:
                preferred.append((10, name))

    preferred = sorted(preferred)
    logger.info(f"preferred interfaces: {preferred}")
    interface = preferred[0][1]
    return interface



import socket
import struct
from asyncio import BaseTransport
from collections import deque
from typing import Any, Iterator, Tuple

import pytest

from aioartnet import (
    DMX_UNIVERSE_SIZE,
    ArtNetClient,
    ArtNetUniverse,
)
from aioartnet.aio_artnet import ArtNetClientProtocol, DGAddr


def test_universe() -> None:
    ac = ArtNetClient()
    assert str(ArtNetUniverse(4, ac)) == "0:0:4"
    assert str(ArtNetUniverse(0x15, ac)) == "0:1:5"
    assert str(ArtNetUniverse(0x315, ac)) == "3:1:5"
    assert str(ArtNetUniverse(0x7FF, ac)) == "7:15:15"
    assert str(ArtNetUniverse(0xFFF, ac)) == "15:15:15"
    assert str(ArtNetUniverse(0x7FFF, ac)) == "127:15:15"
    with pytest.raises(ValueError):
        ArtNetUniverse(0x8FFF, ac)  # only 128 'nets'


def packet_reader(file: str) -> Iterator[Tuple[float, bytes]]:
    with open(file, "rb") as f:
        magic, verMaj, verMin, snaplen, netw = struct.unpack("<IHH8xII", f.read(24))
        print(f"pcap {file} magic {hex(magic)} ver {verMaj}.{verMin} link layer {netw}")

        # magic written as 0xa1b2c3d4 in native order
        # magic reads as 0xa1b2c3d4 => we are usec, good
        # magic reads as 0xd4c3b2a1 => we are usec, byte-swapped
        # magic reads as 0xa1b23c4d => nanos
        assert magic == 0xA1B2C3D4
        timediv = 1000000.0
        while True:
            hdr = f.read(16)
            if len(hdr) == 0:
                return
            tsec, tusec, filesz, wiresz = struct.unpack("<IIII", hdr)
            time = tsec + tusec / timediv
            # print(f" pkt {time} {filesz} {wiresz}")
            packet = f.read(filesz)
            yield time, packet


class MockTransport(BaseTransport):
    def __init__(self) -> None:
        self.sent: list[tuple[bytes, DGAddr]] = []

    def get_extra_info(self, name: str, default: Any = None) -> Any:
        return None

    def sendto(self, data: bytes, addr: DGAddr) -> None:
        self.sent.append((data, addr))


def test_artnet_poll_reply() -> None:
    # play a short pcap recording of Art-Net polls & replies then inspect the resulting data
    # fake client.connect being called by manually building the protocol
    client = ArtNetClient(interface="dummy")
    client.broadcast_ip = "10.10.10.255"
    client.unicast_ip = "10.10.10.10"

    proto = ArtNetClientProtocol(client)
    transport = MockTransport()
    proto.connection_made(transport)

    for _, pkt in packet_reader("test/artnet-nodes.pcap"):
        udp = pkt[42:]
        ip = socket.inet_ntoa(pkt[26:30])
        (port,) = struct.unpack(">H", pkt[34:36])
        print(f"UDP ip {ip}:{port} data {udp!r}")
        # package up the sending address as a tuple like asyncio
        proto.datagram_received(udp, (ip, port))

    assert len(client.nodes) == 2
    # assert len(client.universes) == 4
    assert list(map(str, client.universes.values())) == [
        "0:0:0",
        "0:0:1",
        "0:0:2",
        "0:0:3",
        "0:0:8",
    ]
    # Note that 0:0:8 is being broadcasted to without the node (QLC+) listing the port
    # in its node output port configuration
    assert str(client.nodes[3724650688]) == "ArtNetNode<DMX Monitor,192.168.1.222:6454>"
    assert str(client.nodes[3439438016]) == "ArtNetNode<QLC+,192.168.1.205:6454>"

    # last publisher seq stored by (address,physicalport)
    assert client.universes[8].last_data[0] == 0
    assert client.universes[8].last_data[1] == 0x70
    assert client.universes[8].last_data[2] == 0x94
    assert client.universes[8].publisherseq == {(("192.168.1.205", 6454), 0): 20}

    assert client.universes[2].last_data[1] == 0
    assert client.universes[2].publisherseq == {(("192.168.1.205", 6454), 0): 85}

    # DMX Monitor for iPhone binds from page 1, identifies as a desk
    assert client.nodes[3724650688]._portBinds == {1: []}
    assert client.nodes[3724650688].style == 1

    # QLC binds from page 0, identifies as a node
    assert len(client.nodes[3439438016]._portBinds) == 1
    ports = client.nodes[3439438016]._portBinds[0]
    assert list(map(str, ports)) == [
        "Port<Output,DMX,0:0:0>",
        "Port<Output,DMX,0:0:1>",
        "Port<Output,DMX,0:0:2>",
        "Port<Output,DMX,0:0:3>",
    ]
    assert client.nodes[3439438016].style == 0

    # our node should have replied to the poll
    assert len(transport.sent) == 1
    pollreply, addr = transport.sent[0]
    assert addr == (client.broadcast_ip, 6454)
    assert len(pollreply) == 239

    # because we don't *see* our own poll reply so far in the test
    # (as it's in the MockTransport), our ArtNetNode
    # has not been created yet. Release it now and check
    proto.datagram_received(pollreply, addr)
    assert len(client.nodes) == 3
    nn = client.nodes[168430090]
    assert nn._portBinds == {1: []}
    assert str(nn) == "ArtNetNode<aioartnet,10.10.10.10:6454>"


# all connected protocols recieved each others messages
# note messages are dispatched inline, so the stack is re-entrent in ways
# that a real network is not.
#  ie. a send can trigger a rx that triggers a send that arrives in the
#    middle of the original send. normally event loops don't do this.
class BroadcastTransport(BaseTransport):
    def __init__(self, protocols: list[ArtNetClientProtocol] = []) -> None:
        self.protos = list(protocols)
        self.pending: deque[Tuple[bytes, Any]] = deque()
        for p in self.protos:
            p.connection_made(self)

    def connect_protocol(self, protocol: ArtNetClientProtocol) -> None:
        self.protos.append(protocol)

    def get_extra_info(self, name: str, default: Any = None) -> Any:
        return None

    def sendto(self, data: bytes, addr: DGAddr) -> None:
        self.pending.append((data, addr))

    def drain(self) -> None:
        while self.pending:
            msg = self.pending.popleft()
            for p in self.protos:
                p.datagram_received(*msg)


@pytest.mark.asyncio
async def test_artnet_back_to_back_nodes() -> None:
    # use two instances of our client linked by a mock transport to test
    # port and node detection

    clA = ArtNetClient(interface="dummy", portName="alpha")
    clA.broadcast_ip = "10.10.10.255"
    clA.unicast_ip = "10.10.10.10"

    clB = ArtNetClient(interface="dummy", portName="bravo")
    clB.broadcast_ip = "10.10.10.255"
    clB.unicast_ip = "10.10.10.2"

    protoA = ArtNetClientProtocol(clA)
    protoB = ArtNetClientProtocol(clB)
    transport = BroadcastTransport([protoA, protoB])

    # send, then flush the poll/reply packets
    protoA._send_art_poll()
    transport.drain()

    assert len(clA.nodes) == 2
    assert len(clB.nodes) == 2
    assert (
        str(list(clA.nodes.values()))
        == "[ArtNetNode<alpha,10.10.10.10:6454>, ArtNetNode<bravo,10.10.10.2:6454>]"
    )

    # when a client has a property modified, it automatically sends an unsolicited PollReply
    clB.portName = "charlie"
    assert len(transport.pending) == 1
    transport.drain()

    assert (
        str(list(clA.nodes.values()))
        == "[ArtNetNode<alpha,10.10.10.10:6454>, ArtNetNode<charlie,10.10.10.2:6454>]"
    )
    assert (
        str(list(clB.nodes.values()))
        == "[ArtNetNode<alpha,10.10.10.10:6454>, ArtNetNode<charlie,10.10.10.2:6454>]"
    )


@pytest.mark.asyncio
async def test_ports() -> None:
    # use one instance of client with a mock loopback transport to test
    # port and node detection

    clA = ArtNetClient(interface="dummy", portName="alpha")
    clA.broadcast_ip = "10.10.10.255"
    clA.unicast_ip = "10.10.10.10"
    clA.set_port_config("1:0:7", is_input=True)

    protoA = ArtNetClientProtocol(clA)
    transport = BroadcastTransport([protoA])

    # send, then flush the poll/reply packets
    protoA._send_art_poll()
    transport.drain()

    assert len(clA.nodes) == 1
    assert len(clA.ports) == 1
    assert str(clA._portBinds) == "{1: [Port<Input,DMX,1:0:7>]}"

    # check the *recieved* view of the same packets match
    assert str(list(clA.nodes.values())[0].ports) == "[Port<Input,DMX,1:0:7>]"
    assert list(clA.universes.keys()) == [263]
    assert str(clA.universes[263].publishers) == "[ArtNetNode<alpha,10.10.10.10:6454>]"
    assert clA.universes[263].subscribers == []

    # disable existing, add an output port
    clA.set_port_config("1:0:7")
    clA.set_port_config("0:1:8", is_output=True)

    transport.drain()
    assert str(clA._portBinds) == "{1: [Port<Output,DMX,0:1:8>]}"
    assert str(list(clA.nodes.values())[0].ports) == "[Port<Output,DMX,0:1:8>]"

    assert clA.universes[263].publishers == []
    assert clA.universes[263].subscribers == []
    print(clA.universes)
    assert clA.universes[24].publishers == []
    assert str(clA.universes[24].subscribers) == "[ArtNetNode<alpha,10.10.10.10:6454>]"

    # two ports active at once
    clA.set_port_config("0:1:9", is_input=True)
    transport.drain()
    assert clA.universes[24].publishers == []
    assert str(clA.universes[24].subscribers) == "[ArtNetNode<alpha,10.10.10.10:6454>]"
    assert str(clA.universes[25].publishers) == "[ArtNetNode<alpha,10.10.10.10:6454>]"
    assert str(clA.universes[25].subscribers) == "[]"


@pytest.mark.asyncio
async def test_dmx_tx_rx() -> None:
    # use two instances of our client linked by a mock transport to test
    # port and node detection

    clA = ArtNetClient(interface="dummy", portName="alpha")
    clA.broadcast_ip = "10.10.10.255"
    clA.unicast_ip = "10.10.10.10"

    clB = ArtNetClient(interface="dummy", portName="bravo")
    clB.broadcast_ip = "10.10.10.255"
    clB.unicast_ip = "10.10.10.2"

    protoA = ArtNetClientProtocol(clA)
    protoB = ArtNetClientProtocol(clB)
    transport = BroadcastTransport([protoA, protoB])

    utx = clA.set_port_config("1:0:7", is_input=True)
    urx = clB.set_port_config("1:0:7", is_output=True)

    assert urx.get_dmx() == bytes(DMX_UNIVERSE_SIZE)

    # send, then flush the poll/reply packets
    protoA._send_art_poll()
    transport.drain()

    test_pattern = bytes(list(range(128)) * 4)
    clA.set_dmx(utx, test_pattern)
    assert len(transport.pending) == 1
    transport.drain()

    assert urx.get_dmx() == test_pattern



import ctypes.util
import logging
import os
import socket
from ctypes import (
    CDLL,
    POINTER,
    Structure,
    c_char_p,
    c_int,
    c_uint,
    c_uint8,
    c_void_p,
    pointer,
)
from sys import platform
from typing import Any, Dict, List, Optional

# return MAC addresses under our own constant, because
# on macos socket.AF_PACKET is not defined, and the value
# differs by platform
AF_PACKET = 17

if platform == "linux" or platform == "linux2":

    class Sockaddr(ctypes.Structure):
        _fields_ = [
            ("sa_family", ctypes.c_uint16),
            ("sa_data", ctypes.c_uint8 * (2 + 4 + 16)),
        ]

    ipv4_addr_data_offset = 2

elif platform == "darwin":

    class Sockaddr(Structure):  # type: ignore [no-redef]
        _fields_ = [
            ("sa_len", c_uint8),
            ("sa_family", c_uint8),
            ("sa_data", c_uint8 * 24),
        ]

    ipv4_addr_data_offset = 2
else:
    raise ValueError("Unsupported platform")


class Ifaddrs(Structure):
    pass


Ifaddrs._fields_ = [
    ("ifa_next", POINTER(Ifaddrs)),
    ("ifa_name", c_char_p),
    ("ifa_flags", c_uint),
    ("ifa_addr", POINTER(Sockaddr)),
    ("ifa_netmask", POINTER(Sockaddr)),
    ("ifa_broadaddr", POINTER(Sockaddr)),
    ("ifa_data", c_void_p),
]


def getifaddrs(
    ifname: Optional[str] = None, family: Optional[int] = None
) -> List[Dict[str, Any]]:
    libc = CDLL(
        ctypes.util.find_library("socket" if os.uname()[0] == "SunOS" else "c"),
        use_errno=True,
    )
    libc.getifaddrs.restype = c_int
    ifaddr_p = pointer(Ifaddrs())
    ret = libc.getifaddrs(pointer((ifaddr_p)))
    if ret != 0:
        raise ValueError("getifaddrs nonzero return code")
    addrs = []
    head = ifaddr_p
    while ifaddr_p:
        name = str(ifaddr_p.contents.ifa_name.decode())
        netmask = None
        broadaddr = None
        addr = None
        fam = ifaddr_p.contents.ifa_addr.contents.sa_family
        d = {"name": name, "family": fam}
        if ifaddr_p.contents.ifa_broadaddr:
            broadaddr = bytes(ifaddr_p.contents.ifa_broadaddr.contents.sa_data)
        if ifaddr_p.contents.ifa_netmask:
            netmask = bytes(ifaddr_p.contents.ifa_netmask.contents.sa_data)
        if ifaddr_p.contents.ifa_addr:
            addr = bytes(ifaddr_p.contents.ifa_addr.contents.sa_data)

        if fam == socket.AF_INET:
            if addr:
                d["addr"] = socket.inet_ntoa(addr[2:6])
            if netmask:
                d["netmask"] = socket.inet_ntoa(netmask[2:6])
            if broadaddr:
                d["broadaddr"] = socket.inet_ntoa(broadaddr[2:6])
        elif fam == socket.AF_INET6:
            if addr:
                d["addr"] = socket.inet_ntop(fam, addr[6:22])
            if netmask:
                d["netmask"] = socket.inet_ntop(fam, netmask[6:22])
        elif fam == 17:  # linux MAC addr
            if addr:
                d["addr"] = addr[10:16].hex()
            if broadaddr:
                d["broadaddr"] = broadaddr[10:16].hex()
            fam = AF_PACKET
        elif fam == 18:  # macos MAC addr
            if addr:
                d["addr"] = addr[9:15].hex()
            fam = AF_PACKET
        logging.debug(f"getifaddrs {d}")

        ifaddr_p = ifaddr_p.contents.ifa_next

        if ifname is None or ifname in name:
            if family is None or fam == family:
                addrs.append(d)

    libc.freeifaddrs(head)
    return addrs


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    print(getifaddrs())



from unittest.mock import Mock

import pytest

from aioartnet.console import Engine, Interpreter


@pytest.mark.asyncio
async def test_console() -> None:
    set_dmx = Mock()
    engine = Engine(set_dmx, universe_size=20)
    interpreter = Interpreter(engine)

    exp = bytearray(20)
    await interpreter.on_cmd("live on")
    set_dmx.assert_called_with(exp)
    assert engine.cues == []

    await interpreter.on_cmd("chan 1 at 50")
    exp[0] = 128
    set_dmx.assert_called_with(exp)

    await interpreter.on_cmd("chan 10 thru 20 at 20")
    exp[9:19] = bytes([51] * 10)
    set_dmx.assert_called_with(exp)

    # saving clears the live edits
    await interpreter.on_cmd("RECORD CUE 1")
    assert len(engine.cues) == 1
    zeros = bytearray(20)
    set_dmx.assert_called_with(zeros)

    await interpreter.on_cmd("RECORD CUE 2 HOLD 4")
    await interpreter.on_cmd("RECORD CUE 3 HOLD 2 FADE 1")
    assert len(engine.cues) == 3

    # await interpreter.on_cmd("CUE 3 LABEL \"Scene 1 blackout\"")
    assert engine.active_cue is None
    await interpreter.on_cmd("GO")
    assert engine.active_cue == 0

    set_dmx.assert_called_with(exp)

    # await interpreter.on_cmd("STATE")
    # await interpreter.on_cmd("LIST")


@pytest.mark.asyncio
async def test_console_loop() -> None:
    set_dmx = Mock()
    engine = Engine(set_dmx, universe_size=20)
    interpreter = Interpreter(engine)
    cmds = "ch 1 at f,record cue 1,ch 2 at f,record cue 2,ch 3 at f,record cue 3,go,go,go,go"
    for cmd in cmds.split(","):
        await interpreter.on_cmd(cmd)
    assert engine.active_cue == 0



import argparse
import asyncio
import logging
import os
import time
import traceback
from dataclasses import dataclass, field
from enum import Enum, auto
from typing import Any, Callable, Sequence

from prompt_toolkit.history import FileHistory
from prompt_toolkit.patch_stdout import patch_stdout
from prompt_toolkit.shortcuts import PromptSession

from . import DMX_UNIVERSE_SIZE, ArtNetClient


class FadeState(Enum):
    OFF = auto()
    FADE_IN = auto()
    HOLD = auto()
    FADE_OUT = auto()


@dataclass
class ChannelIntensity:
    channel: int
    intensity: int


@dataclass
class Cue:
    name: str
    fade_in: float
    hold: float
    fade_out: float
    channels: list[ChannelIntensity] = field(default_factory=list)


@dataclass
class Submaster:
    intensity: float
    name: str
    channels: list[ChannelIntensity] = field(default_factory=list)


@dataclass
class ActiveCue:
    cue: Cue
    since: float
    state: FadeState = FadeState.FADE_IN

    def get_update_intensity(self, time: float) -> float:
        cue = self.cue
        uptime = time - self.since
        if uptime < cue.fade_in:
            self.state = FadeState.FADE_IN
            return uptime / cue.fade_in
        elif uptime < cue.fade_in + cue.hold:
            self.state = FadeState.HOLD
            return 1.0
        elif uptime < cue.fade_in + cue.hold + cue.fade_out:
            self.state = FadeState.FADE_OUT
            return 1.0 - ((uptime - cue.fade_in - cue.hold) / cue.fade_out)
        else:
            self.state = FadeState.OFF
            return 0.0


def apply_ci(data: bytearray, edits: Sequence[ChannelIntensity], scale: float) -> None:
    for e in edits:
        data[e.channel] = max(0, min(255, int(e.intensity * scale)))


class Engine:
    def __init__(
        self, handler: Callable[[bytes], None], universe_size: int = DMX_UNIVERSE_SIZE
    ):
        self.universe_size = universe_size
        self.last_poll: float = 0
        self.cues: list[Cue] = []
        # an index into cues[] of our current playback position, or None if stopped
        self.active_cue: int | None = None
        self.subs: list[Submaster] = []
        self.live = bytearray(universe_size)
        self.handler = handler
        self.loop = False
        self.edits: list[ChannelIntensity] = []
        self.live_edit = False
        self.tickhz = 10
        self._active_cues: list[ActiveCue] = []

    async def poll(self, time: float) -> None:
        # advance time
        self.last_poll = max(self.last_poll, time)

        live = bytearray(self.universe_size)
        # every cue is either off, fade-in, hold, or fade-out
        for ac in self._active_cues:
            intensity = ac.get_update_intensity(self.last_poll)
            apply_ci(live, ac.cue.channels, scale=intensity)

        for ac in self._active_cues:
            if ac.state == FadeState.OFF:
                print(f"finished {ac}")

        self._active_cues[:] = [
            x for x in self._active_cues if x.state != FadeState.OFF
        ]

        # mix in any non-zero submasters

        # add live edits
        if self.live_edit:
            apply_ci(live, self.edits, 1.0)

        # print(f'calling handler {self.handler} with {self.live}')
        self.handler(live)

    async def go_relative(self, change: int) -> None:
        if len(self.cues) == 0:
            self.active_cue = None
            return
        # special handling for resuming/starting first cue
        if self.active_cue is None:
            next_cue = 0 if change == 1 else len(self.cues)
        else:
            next_cue = self.active_cue + (1 if change == 1 else -1)
        # if current queue is held, nudge it to fade-out
        next_cue = next_cue % len(self.cues)
        self.activate(self.cues[next_cue])
        self.active_cue = next_cue

    def activate(self, cue: Cue) -> None:
        # activates a cue, with the start time of our last poll
        self._active_cues.append(ac := ActiveCue(since=self.last_poll, cue=cue))
        print(f"activated {ac}")

    async def stop(self) -> None:
        self.active_cue = None

    async def go_absolute(self, cue_num: int) -> None:
        # cc = self.cues[self.active_cue]
        print(cue_num)

    async def clear_edits(self) -> None:
        self.edits = []
        await self.poll(0)

    def add_edit(self, edit: ChannelIntensity) -> None:
        ci = self.edits
        for i in range(len(ci)):
            if ci[i].channel == edit.channel:
                ci[i] = edit
                break
        else:
            ci.append(edit)

    async def start_ticking(self) -> asyncio.Task[None]:
        self._tick_task = asyncio.create_task(self._tick())
        return self._tick_task

    async def _tick(self) -> None:
        while True:
            await asyncio.sleep(0.05)
            await self.poll(time.time())


def parse_intensity(level: str) -> int:
    if level.upper() in ("F", "FL", "FULL"):
        return 255
    elif level.upper() in ("Z", "ZERO"):
        return 0
    elif level.upper() in ("H", "HALF"):
        return 128
    elif level.lower().startswith("0x"):
        return min(255, int(level[2:], base=16))
    else:
        return min(255, int((int(level) / 100.0) * 256))


def parse_user_index(value: str, items: Sequence[Any], extend: bool = False) -> int:
    # python indexes from 0 to len(items)-1
    # we allow 'user' indexing from 1 to len(items) for overwrites, and
    # from 1 to len(items)+1 if we are doing a write operation to extend the
    # collection
    cn = int(value) - 1
    if cn < 0:
        raise ValueError("numbering starts from 1")
    if cn > len(items) + {True: 1, False: 0}[extend]:
        raise ValueError(f"highest existing item is {len(items)}")
    return cn


def parse_duration(value: str) -> int:
    return int(value)


# interprets simple commands and formats the responses
class Interpreter:
    def __init__(self, engine: Engine):
        self.engine = engine

    async def on_cmd(self, cmd: str) -> str:
        match cmd.lower().split():
            case ["live", ("on" | "off") as mode]:
                self.engine.live_edit = mode == "on"
            case ["loop", ("on" | "off") as mode]:
                self.engine.loop = mode == "on"
            case ["chan" | "ch", chan, ("at" | "@"), level]:
                intensity = parse_intensity(level)
                self.engine.add_edit(
                    ChannelIntensity(channel=int(chan) - 1, intensity=intensity)
                )
            case ["chan" | "ch", chan_from, "thru", chan_to, "at", level]:
                intensity = parse_intensity(level)
                for i in range(int(chan_from), int(chan_to)):
                    self.engine.add_edit(
                        ChannelIntensity(channel=i - 1, intensity=intensity)
                    )
            case ["sub" | "submaster", chan, ("at" | "@"), level]:
                intensity = parse_intensity(level)
                parse_user_index(chan, self.engine.subs, extend=False)
                # self.engine.subs[]
            case ["record", ("cue" | "sub") as target, cue_num, *args]:
                # record cue 1 time 3
                fade_in = 0
                fade_out = 0
                hold = 1
                while args:
                    match args:
                        case ["fade", duration, *more]:
                            fade_in = parse_duration(duration)
                            fade_out = fade_in
                        case ["fade_in", duration, *more]:
                            fade_in = parse_duration(duration)
                        case ["fade_out", duration, *more]:
                            fade_out = parse_duration(duration)
                        case ["hold", duration, *more]:
                            hold = parse_duration(duration)
                        case _:
                            raise ValueError(f"unknown record args {args}")
                    args = more

                if target == "cue":
                    cn = parse_user_index(cue_num, self.engine.cues, extend=True)
                    cue = Cue(
                        name="",
                        fade_in=fade_in,
                        hold=hold,
                        fade_out=fade_out,
                        channels=list(self.engine.edits),
                    )
                    print(f"inserting {cue} as cue {cn}")
                    if cn == len(self.engine.cues):
                        self.engine.cues.append(cue)
                    else:
                        self.engine.cues[cn] = cue
                    await self.engine.clear_edits()
                else:
                    pass
            # update cue 1
            # update
            case ["edits" | "edit" | "dirty"]:
                print(self.engine.edits)
            case ["go"]:
                await self.engine.go_relative(1)
            case ["back"]:
                await self.engine.go_relative(-1)
            case ["go", cue_num]:
                cn = int(cue_num) - 1
                await self.engine.go_absolute(cn)
            case ["cue", cue_num, "go"]:
                raise ValueError("NYI")
            case ["clear"]:
                await self.engine.clear_edits()
            case ["list"]:
                if self.engine.cues:
                    for idx, cue in enumerate(self.engine.cues):
                        print(f"cue {idx + 1:03} {cue}")
                else:
                    print("No cues")
                if self.engine.subs:
                    for idx, sub in enumerate(self.engine.subs):
                        print(f"sub {idx + 1:03} {sub}")
                else:
                    print("No submasters")
            case ["tickhz", hz]:
                self.engine.tickhz = int(hz)
            case _:
                raise ValueError(f"Unknown command: {cmd}")

        await self.engine.poll(0)

        return ""

    async def load_commands(self, filename: str) -> None:
        with open(filename, "r") as file:
            for line in file.read():
                await self.on_cmd(line.strip())


async def main(client: ArtNetClient, engine: Engine, interpreter: Interpreter) -> None:
    history = FileHistory(os.path.expanduser("~/.aioartnet-console-history"))
    session: PromptSession[str] = PromptSession("> ", history=history)

    await client.connect()
    await engine.start_ticking()

    # Run echo loop. Read text from stdin, and reply it back.
    while True:
        try:
            inp = await session.prompt_async("> ")
            if inp == "":
                continue
            output = await interpreter.on_cmd(inp)
            if output is None or output == "":
                continue
            print(output)

        except KeyboardInterrupt:
            return
        except EOFError:
            break
        except Exception:
            traceback.print_exc(limit=-2)
    return None


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="aioartnet console",
        description="Replica vintage console",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("-v", "--verbose", action="store_true")
    parser.add_argument("-i", "--interface")
    parser.add_argument("-n", "--portName")
    parser.add_argument(
        "-u", "--universe", help="Art-Net universe to output", default="0:0:0"
    )
    parser.add_argument("-f", "--file", help="read commands from file")
    args = parser.parse_args()

    level = {False: logging.INFO, True: logging.DEBUG}[args.verbose]
    logging.basicConfig(level=level)

    # setup art-net input universe (for our output)
    kwargs = {}
    if args.interface:
        kwargs["interface"] = args.interface
    if args.portName:
        kwargs["portName"] = args.portName
    client = ArtNetClient(**kwargs)
    u1 = client.set_port_config(args.universe, is_input=True)

    # setup console engine/interpreter wired to our universe
    engine = Engine(u1.set_dmx)
    interpreter = Interpreter(engine)

    # TODO: needs await as cmds are async?
    # if args.file:
    #    interpreter.load_commands(args.file)

    with patch_stdout():
        asyncio.run(main(client, engine, interpreter))



import argparse
import asyncio
import logging

from . import ArtNetClient


async def main(client: ArtNetClient) -> None:
    await client.connect()
    # u5 = client.set_port_config("0:0:5", isoutput=True)

    while True:
        await asyncio.sleep(5)
        print("nodes:")
        for n, node in client.nodes.items():
            print(f" {node!r: <60} {node.ports}")
        print("universes:")
        for univ in client.universes.values():
            print(f" {univ} pubs:{univ.publishers} subs:{univ.subscribers}")

        # print(u5.last_data[0:20].hex())


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="aioartnet cli",
        description="View Art-Net nodes and universes",
    )
    parser.add_argument("-v", "--verbose", action="store_true")
    parser.add_argument("-i", "--interface")
    parser.add_argument("-n", "--portName")
    parser.add_argument("-p", "--publish", action="store_true")
    args = parser.parse_args()

    level = {False: logging.INFO, True: logging.DEBUG}[args.verbose]
    logging.basicConfig(level=level)

    kwargs = {}
    if args.interface:
        kwargs["interface"] = args.interface
    if args.portName:
        kwargs["portName"] = args.portName
    client = ArtNetClient(**kwargs)
    if args.publish:
        u1 = client.set_port_config("0:0:0", is_input=True)
        u1.set_dmx(bytes(list(range(128)) * 4))
    asyncio.run(main(client))
    asyncio.get_event_loop().run_forever()
